#!python3

# split.py - Split files into folders, based on their names.
#
# Copyright(C) 2024, Ian Michael Dunmore
#
# License: https://github.com/idunmore/AtariTools/blob/master/LICENSE

# Native Python Modules
import itertools
import pathlib
import os.path
import shutil
import string
from typing import Self, Iterator

# 3rd Party/External Modules
import click

# Local Application Modules
from .style import C

# Constants

# Error Messages and Command Result Exit Codes
ERROR_TEXT = f'{C.Err}Error: {C.Off}'
ERROR = 1
SUCCESS = 0

# Verbosity Values
SILENT = 0
PROGRESS = 1
VERBOSE = 2

# Default Argument Values
DEFAULT_MAX_FILES_PER_FOLDER = 250
DEFAULT_MIN_FILES_PER_FOLDER = 1
DEFAULT_PRESERVE_GROUPING = False

# File and Path Patterns
SOURCE_FILE_PATTERN = '**/[!.]*.*'
NUMERIC_FOLDER_NAME = '0-9'

BAND_RANGE_SEPARATOR = '-'
BAND_SEPARATOR = ','
FOLDER_RANGE_SEPARATOR = '-'

# File and Folder Structures
class Folder(list):
    '''A list of files, representing a single folder.'''
    def __init__(self: Self, name: str = ''):
        super().__init__()
        self._name = name

    @property
    def name(self: Self) -> str:
        return self._name
    
    @name.setter
    def name(self: Self, name: str):
        self._name = name

    @property
    def first_filename(self: Self) -> str:
        '''Returns the name of the first file in the folder.'''
        return '' if len(self) == 0 else self[0].name 

    @property
    def last_filename(self: Self) -> str:
        '''Returns the name of the last file in the folder.'''
        return '' if len(self) == 0 else self[len(self) - 1].name
    
    @property
    def common_prefix(self: Self) -> str:
        '''Returns the common prefix of all files in the folder.'''
        return os.path.commonprefix([file.name for file in self])
    
class Folders(list):
    '''A list of folders, representing the entire file structure.'''
    def __init__(self: Self):
        super().__init__()      
    
# Splitters

class Splitter:
    '''Base class for all file splitters.'''
    def __init__(self: Self, files: list):
        self._files = files
        files.sort(key=lambda x : x.name.lower())

    def split(self: Self) -> Folders:
        '''Splits the files into folders.'''
        return Folders()

    @property
    def folders(self: Self):
        return None
    
class SimpleSplit(Splitter):
    '''Splits files individual folders, based on the file's first character.'''
    def __init__(self: Self, files: list):
        super().__init__(files)

    def split(self: Self) -> Folders:
        '''Splits the files into folders, one per unique first character.'''
        key_func = lambda x: x.name[0].lower()
        folder_dict = {key: list(group) for key, group
            in itertools.groupby(self._files, key_func)}
        
        # Create the Folders container, to add individual folders to.
        folders = Folders()

        self._add_numeric_folder(folder_dict, folders)
        self._add_alpha_folders(folder_dict, folders)

        return folders     
    
    def _add_numeric_folder(self: Self, folder_dict: dict, folders: Folders):
        # Group all the numeric prefixes into one folder.
        folder = Folder(NUMERIC_FOLDER_NAME)        
        
        for key in folder_dict.keys():
            if key in string.digits:
                for file in folder_dict[key]:
                    folder.append(file)
        folders.append(folder)

    def _add_alpha_folders(self:Self, folder_dict: dict, folders: Folders):
        '''Add all non-numeric folders to the Folders container.'''
        for key in folder_dict.keys():
            if key in string.ascii_lowercase:
                folder = Folder(key.upper())
                for file in folder_dict[key]:
                    folder.append(file)
                folders.append(folder)
    
    @property
    def folders(self: Self):
        return None

class BandSplit(Splitter):
    '''Splits files folders, based on groups of the file's first character.'''
    def __init__(self: Self, files: list):
        super().__init__(files)

    def split(self: Self, bands: list) -> Folders:
        key_func = lambda x: x.name[0].lower()
        folder_dict = {key: list(group) for key, group
            in itertools.groupby(self._files, key_func)}
        
        folders = Folders()
        self._add_folder_bands(folder_dict, folders, bands)

        return folders
    
    def _add_folder_bands(
        self: Self, folder_dict: dict, folders: Folders, bands: list):
            # Go one band at a time ...
            for band in bands:
                start, end = self._get_band_extents(band.lower())
                # ... creating a new folder for each band ...
                folder = Folder(band.upper())
                # ... then for every character prefix in the band ...
                for char in char_range(start, end):
                    # ... find the matching folders in the overall list ...                   
                    for key in folder_dict.keys():
                        if key[0] == char:
                            # ... and all all of its files to the folder.
                            for file in folder_dict[key]:
                                folder.append(file)
                # Add this folder, which contains all files for this band
                # to the overall list of folders.
                folders.append(folder)                
                
    def _get_band_extents(self: Self, band: str) -> tuple[str, str]:
        '''Returns the first and last characters (range) of a band.'''
        extents = band.split(BAND_RANGE_SEPARATOR)
        if len(extents) != 2:
            raise ValueError(
                f'{ERROR_TEXT}Invalid band: "{band}" - '
                f'expected format is "A-E", "A-Z", or "0-9"')
        
        start = extents[0]
        end = extents[1]        
        if start >= end:
            raise ValueError(
                f'{ERROR_TEXT}Invalid band: "{band}" - '
                f'end ({end}) must be greater than start ({start})')
        
        return (start, end)

class MaxFileSplit(Splitter):
    '''Splits files individual folders, based on the file's first character.'''
    def __init__(self: Self, files: list):
        super().__init__(files)

    def split(self: Self, max_files_per_folder: int, group: bool) -> Folders:
        '''Splits the files into folders, with a maximum number of files per
        folder.  Optionally, tries to keep files with like-prefixes together,
        which may result in fewer files per folder.'''
        total_files = len(self._files)
        current_file = 0
        folder_file_count = 0
        folders = Folders()
        folder = Folder()
        folders.append(folder)

        # Process all files ... one at a time, in order ...
        while(current_file < total_files):
            # ... if there's still room in the folder, add the current file ...
            if folder_file_count < max_files_per_folder:
                folder.append(self._files[current_file])
                folder_file_count += 1
                current_file += 1
            else:
                # Folder is full, so we need another; first adjust the current
                # folder (which may remove files) to group like-prefixes.
                if group:
                    current_file = self._adjust_folder(folder, current_file)
                
                # Now we can start a new folder.
                folder = Folder()
                folders.append(folder)
                folder_file_count = 0

        self._set_folder_names(folders)
        return folders

    def _adjust_folder(self: Self, folder: Folder, current_file: int) -> int:
        # Adjust folder content to keep like-prefixes together as much as
        # possible.  For example, don't split 'AA*' files into two folders;
        # instead stop at 'AAA*' and start a new folder for # 'AAB*' files.
        prefix_length = len(folder.common_prefix) + 1
        last_prefix = self._files[current_file - 1].name[:prefix_length]
        current_prefix = self._files[current_file].name[:prefix_length]

        # Nothing to do unless the prefixes of the first and last files in the
        # folder have the same prefix.
        if last_prefix == current_prefix:
            # Starting with the PREVIOUS file ...
            current_file -= 1
            while(current_file > 0):
                # ... remove files from the folder until prefixes don't match.
                if (self._files[current_file].name[:prefix_length] ==
                    last_prefix ):
                    folder.pop()
                    current_file -= 1
                else:
                    break

            # Don't add the LAST file to both folders!
            current_file += 1

        return current_file

    def _set_folder_names(self: Self, folders: Folders):
        '''Sets the names of the folders based on their contents.'''
        # Setup the start of the name for the first folder ...           
        self._set_first_folder_name_start(folders)    
        # We do the start of all ours folder's name/ranges first ...
        self._set_folder_name_start(folders)     
        # ... as the end name is dependent on the start name/range.   
        self._set_folder_name_end(folders)    
         # Unique processing for the LAST folder.        
        self._set_last_folder_name_end(folders)       

    def _set_first_folder_name_start(self: Self, folders: Folders):
        '''Sets the start of the first folder's name.'''
        # The FIRST folder's start prefix is ALWAYS just the first character
        # of its FIRST file name.            
        folders[0].name = folders[0].first_filename[0].upper()

    def _set_folder_name_start(self: Self, folders: Folders):
        '''Sets the start name range for each folder.'''
        # We've already done the first folder, so continue from the second.
        current_folder_index = 1
        
        # We are just doing the START name for each folder.       
        num_folders = len(folders)
        while(current_folder_index < num_folders):
            last_folder = folders[current_folder_index - 1]
            current_folder = folders[current_folder_index]

            # The start of the folder name is the name of its first filename
            # UPTO the point that it first differs from its last filename.
            start_length = first_different_character_index(
                current_folder.first_filename,
                last_folder.last_filename)            
            start = current_folder.first_filename[:start_length + 1]
            current_folder.name = f'{start}'.upper()

            current_folder_index += 1

    def _set_folder_name_end(self: Self, folders: Folders):        
        num_folders = len(folders)           
        current_folder_index = 0
        # Do all folders, except the last one - which get special handling.
        while(current_folder_index < num_folders - 1):
            current_folder = folders[current_folder_index]
            next_folder = folders[current_folder_index + 1]            

            # The simple end of the folder name is name of its last filename
            # UPTO the point that it first differs from the next folder's
            # first filename.                  
            end_length = first_different_character_index(
                current_folder.last_filename,
                next_folder.first_filename)
            
            # If the current (start) name of the folder is longer than the end
            # AND it is s different character, then we can shorted end by one
            # character so we're not overqualifying the end name.
            if (len(current_folder.name) > end_length and
                current_folder.first_filename[0] ==
                current_folder.last_filename[0]):
                    end_length = len(current_folder.name) - 1        

            end = current_folder.last_filename[:end_length + 1].upper()

            # Add end to the folder name ONLY if it is different to start (name)
            if current_folder.name != end:
                current_folder.name += f'{FOLDER_RANGE_SEPARATOR}{end}'.upper()
            
            current_folder_index += 1       

    def _set_last_folder_name_end(self: Self, folders: Folders):
        '''Sets the end prefix of the last folder.'''
        # Get the last folder in the Folders collection.
        last_folder = folders[len(folders) - 1]
        start = last_folder.name
          
        if start[0] == last_folder.last_filename[0]:
            # The last file in the folder has the same prefix as the first, so
            # we need to make the end as long as the start.
            end = last_folder.last_filename[0:len(start)]
        else:
            # The last file in the folder has a different prefix than the
            # first, so we only need the first letter of the last file's name
            # (as there can be no more files/divisions).
            end = last_folder.last_filename[0]

        # If the start and end range of the folder is the same, then we don't
        # need to show the end (or the separator).
        end = '' if start == end else f'{FOLDER_RANGE_SEPARATOR}{end}'        
        last_folder.name = f'{start}{end}'.upper()

# Command Line Interface

@click.group()
@click.version_option()
def split():
    '''Moves files to organized folder structures.'''
    pass

@split.command('alpha')
@click.option('-c', '--copy', is_flag=True, default=False,
    help= f'Copies files to new partitions (else does nothing)')
@click.option('-v', '--verbosity', type=click.Choice(['0', '1', '2']),
    default='1', show_default=True, help='Status/progress reporting verbosity')
@click.argument('source_path', default='./',
    type=click.Path(exists=True, file_okay=False, dir_okay=True))
@click.argument('dest_path', default='./',
    type=click.Path(exists=False, file_okay=False, dir_okay=True))
def alpha(copy: bool, verbosity: str, source_path: str, dest_path: str) -> int:
    '''Splits source files into smaller, alphabetic/numeric folders.'''    
    # Sanity check input
    validate_paths(source_path, dest_path)

    # Get the files to be split ...
    file_list = build_source_file_list(source_path)
    # ... select the approrpiate splitter ...
    splitter = SimpleSplit(file_list)
    # ... get the split folder/filer structure ...
    folders = splitter.split()
    # ... and then process the files:
    process_splits(folders, pathlib.Path(dest_path), copy, int(verbosity))

@split.command('band')
@click.option('-c', '--copy', is_flag=True, default=False,
    help= f'Copies files to new partitions (else does nothing)')
@click.option('-v', '--verbosity', type=click.Choice(['0', '1', '2']),
    default='1', show_default=True, help='Status/progress reporting verbosity')
@click.option('-b', '--bands', default='0-9,a-e,f-j,k-o,p-t,u-z',
    show_default=True, help='Comma-separated list of bands ("0-9,A-E" etc.)')   
@click.argument('source_path', default='./',
    type=click.Path(exists=True, file_okay=False, dir_okay=True))
@click.argument('dest_path', default='./',
    type=click.Path(exists=False, file_okay=False, dir_okay=True))
def band(copy: bool, verbosity: str, bands: str,
	source_path: str, dest_path: str):
    '''Splits source files into smaller, "bands" (e.g., 0-9, A-E etc.).'''
    
    validate_paths(source_path, dest_path)

    # Get the files to be split ...
    file_list = build_source_file_list(source_path)
    splitter = BandSplit(file_list)
    # ... get the split folder/filer structure ... 
    # ... removing any spaces in the "bands" specification ...  
    bands = bands.replace(' ', '')
    folders = splitter.split(list(bands.split(BAND_SEPARATOR)))
    # ... and then process the files:
    process_splits(folders, pathlib.Path(dest_path), copy, int(verbosity))   

@split.command('max')
@click.option('-c', '--copy', is_flag=True, default=False,
    help= f'Copies files to new partitions (else does nothing)')
@click.option('-g', '--group', is_flag=True, default=DEFAULT_PRESERVE_GROUPING,
    help='Groups files by like prefixes')
@click.option('-m', '--max_files', type=int, show_default=True,
    default=DEFAULT_MAX_FILES_PER_FOLDER,
    help='Maximum # of files per partition')
@click.option('-v', '--verbosity', type=click.Choice(['0', '1', '2']),
    default='1', show_default=True, help='Status/progress reporting verbosity')
@click.argument('source_path', default='./',
    type=click.Path(exists=True, file_okay=False, dir_okay=True))
@click.argument('dest_path', default='./',
    type=click.Path(exists=False, file_okay=False, dir_okay=True))
def max(copy: bool, group: bool, max_files: int, verbosity: str,
        source_path: str, dest_path: str) -> int:
    '''Splits source files into folders, with a max # of files each.
        Optionally, tries to group files with like-prefixes together
        (may result in fewer than the max # of files per folder).'''
    
    # Sanity check input
    validate_paths(source_path, dest_path)

    if max_files < DEFAULT_MIN_FILES_PER_FOLDER:
        click.echo(
            f'{ERROR_TEXT}Maximum number of filers per parition must be '
            f'at least {DEFAULT_MIN_FILES_PER_FOLDER}.', err=True)
        exit(ERROR)   

    file_list = build_source_file_list(source_path)
    splitter = MaxFileSplit(file_list)
    # ... get the split folder/filer structure ...       
    folders = splitter.split(max_files, group)
    # ... and then process the files:
    process_splits(folders, pathlib.Path(dest_path), copy, int(verbosity))
   
# Splitter Processing Functions
def process_splits(folders: Folders, dest_path: pathlib.Path,
    copy: bool, verbosity: int):
    '''Processes Folders and splits the files based on its contents.'''

    # Iterate all folders ...
    if verbosity == PROGRESS:
        # ... showing a progress bar.
        with click.progressbar(folders, label='Processing folders') as bar:
            for folder in bar:
                process_folder(folder, dest_path, copy, verbosity)
    else:
        # ... no progress bar; verbosity sets silent or file-by-file logging
        for folder in folders:
            process_folder(folder, dest_path, copy, verbosity)        

def process_folder(folder: Folder, dest_path: pathlib.Path, copy: bool,
    verbosity: int):
    '''Processes a single folder, copying/moving its files to the destination.'''

    folder_path = dest_path / folder.name    
    echo_v(f'Creating folder: {folder_path}', verbosity)
    if copy:
        # Only create folders if we're in COPY mode.        
        folder_path.mkdir(parents=True, exist_ok=True)

    for file in folder:
        dest_file = folder_path / file.name
        echo_v(f'Copying file: {file} to {dest_file}', verbosity)
        if copy:
            # Only copy files if we're in COPY mode.
            shutil.copy(file, dest_file, follow_symlinks=False)
            
# GENERAL Utility Functions

def validate_paths(source_path, dest_path):
    '''Validates source and dest paths; if not valid exits with error.'''
    if source_path == dest_path:
        click.echo(
            f'{ERROR_TEXT}SOURCE_PATH: "{source_path}" and DEST_PATH: '
            f'"{dest_path}" cannot be the same.', err=True)
        exit(ERROR)

def build_source_file_list(source_path: str) -> list:
    '''Builds the raw list of source files to be partitioned.'''
    source_path = pathlib.Path(source_path)
    # Don't include directories in the list of FILEs to move, as each file will
    # include it's full path anyway.
    source_files = [file for file in list(source_path.glob(SOURCE_FILE_PATTERN))
                    if not file.is_dir()]
    # Sort the files, by name, as they will be grouped alphaetically.
    source_files.sort(key=lambda x : x.name)
    return source_files 

def char_range(start: str, end: str) -> Iterator[str]:
    '''Generates the characters from start to end, inclusive.'''
    for c in range(ord(start), ord(end) + 1):
        yield chr(c)

def first_different_character_index(first: str, second: str) -> int:
    '''Gets index of first character that differs between two strings.'''
    for index, (char1, char2) in enumerate(zip(first, second)):
        if char1 != char2:
            return index
    return -1

def echo_v(message: str, verbosity: int):
    if verbosity == VERBOSE:
        click.echo(message)	

# Run!
if __name__ == '__main__':
    split()


#!python3

# config.py - Atari THE400 Mini Configuration Utility & Functions
#
# Copyright(C) 2024, Ian Michael Dunmore
#
# License: https://github.com/idunmore/AtariTools/blob/master/LICENSE

# Native Python Modules
import pathlib
import shutil

# 3rd Party/External Modules
import click

# Local Application Modules
from .style import C

# Constants

# Error Messages and Command Result Exit Codes
ERROR_TEXT = f'{C.Err}Error: {C.Off}'
ERROR = 1
SUCCESS = 0

# Verbosity Values
SILENT = 0
PROGRESS = 1
VERBOSE = 2

# File Constants
TARGET_PATTERN = '*.*'
TARGET_PATTERN_RECR = '**/*'

# Suffix and Extension NEED to be different due to the way pathlib works;
# path.with_suffix() replaces the last suffix with the new one, where
# path.suffix does NOT return the separating period.
CFG_SUFFIX = '.cfg'
CFG_EXTENSION = 'cfg'

# THE400 Mini CONFIG Constants

EOL = ';'
KEY_VALUE_SEPARATOR = '='

# Atari Media File Extensions
MEDIA_EXTENSIONS = ['atr', 'atx', 'xfd', 'dcm', 'com', 'exe', 'xex', 'cas',
                    'car', 'crt', 'rom', 'bin', 'a52', 'm3u']
FIRST_CARTRIDGE_TYPE = 1
LAST_CARTRIDGE_TYPE = 70

# Emulator Keys & Defaults
KEY_EMULATOR_PREFIX = 'emulator_'
KEY_EMULATOR_MACHINE = 'machine'
KEY_EMULATOR_ARTEFACT = 'artefact'
KEY_EMULATOR_FORCE_PAL = 'force_pal'
DEFAULT_MACHINE = 'XE'
DEFAULT_ARTEFACT = '0'
DEFAULT_FORCE_PAL = 'false'
EMULATOR_SEPARATOR = '-'
BASIC = 'basic'

# Map CLI Model IDs (key) to Config File Machine IDs (value)
MODEL_MACHINE_ID = {'400':'400', '800':'800', 'XL':'800xl', 'XE':'130xe',
                 '5200':'5200'}

# Controller Keys & Defaults
CONTROLLER_DEFAULTS = {'a':'joy_a', 'b':'start', 'x':'option', 'y':'select',
	'up':'joy_up', 'down':'joy_down', 'left':'joy_left', 'right':'joy_right',
	'sl':'joy_a', 'sr':'joy_a', 'menu':'option'}

CONTROLLER_5200_DEFAULTS = {'a':'joy_a', 'b':'5200_start', 'x':'option',
	'y':'5200_pause', 'up':'axis_left_y', 'down':'axis_left_y',
	'left':'axis_left_x', 'right':'axis_left_x', 'sl':'joy_a', 'sr':'joy_a',
	'menu':'option'}

CONTROLLER_PREIX = 'controller_'
CONTROLLER_SEPARATOR = '_'
NUMBER_OF_CONTROLLERS = 4

KEY_INPUT_ANALOG_PREFIX = 'input_analogue_'
KEY_ANALOG_RATE = 'rate'
KEY_ANALOG_CURVE = 'curve'
KEY_ANALOG_AUTO_CENTER = 'autocenter'
DEFAULT_ANALOG_RATE = 100
DEFAULT_ANALOG_CURVE = 'linear'
DEFAULT_AUTO_CENTER = 'off'

# Display Keys & Defaults

KEY_DISPLAY_PREFIX = 'display_'
KEY_DISPLAY_START = 'start_line'
KEY_DISPLAY_HEIGHT = 'height'
KEY_DISPLAY_WIDTH = 'width'
DEFAULT_DISPLAY_START = 20
DEFAULT_DISPLAY_HEIGHT = 200
DEFAULT_DISPLAY_WIDTH = 320

# Command Line Interface

@click.group()
@click.version_option()
def config():
    '''Creates and updates .CFG files for THE400 Mini USB Media games.'''
    pass

@config.command()
@click.option('-m', '--model', default=DEFAULT_MACHINE, show_default=True,
	type=click.Choice(['400', '800', 'XL', 'XE', '5200']),
	help='Atari 8-bit Model to configure for (where XL=800XL and XE=130XE)')
@click.option('-b', '--basic', is_flag=True, default=False, help="Enable BASIC")
@click.option('-vs', '--video_std', type=click.Choice(['NTSC', 'PAL']),
	default="NTSC", show_default=True,
	help='Video Standard (NTSC=60Hz, PAL=50Hz)')
@click.option('-p', '--force_pal', is_flag=True, default=False,
	help='Forces PAL operation (60Hz refresh)')
@click.option('-a', '--artefact', default='0', show_default=True,
	type=click.Choice(['0','1','2','3']), help="Artefacting mode")
@click.option('-s', '--start', default=DEFAULT_DISPLAY_START, show_default=True,
	help='Display Start scanline (lower values start higher on screen)')
@click.option('-h', '--height', default=DEFAULT_DISPLAY_HEIGHT,
    show_default=True, help='Display Height (scanlines)')
@click.option('-w', '--width', default=DEFAULT_DISPLAY_WIDTH, show_default=True,
	help='Display Width (pixels)')
@click.argument('filename', type=click.File('w'), default='-')
def create(filename, model: str, basic: bool, video_std: str, force_pal: bool,
	artefact: str, start: int, height: int, width: int):
    '''Creates .cfg files for Atari THE400 Mini.'''
    # Build the configuration file.
    configuration = build_config(model, basic, video_std, force_pal, artefact,
       start, height, width)
    
    # Write the configuration file to the specified file, or stdout.
    for line in configuration:
        filename.write(line + '\n')
    
    filename.close()
    exit(SUCCESS)    

@config.command()
@click.option('-o', '--overwrite', is_flag=True, default=False,
        help='Overwrite existing .cfg files')
@click.option('-r', '--recurse', is_flag=True, default=False,
    help='Recursively process all files in target directory')
@click.option('-v', '--verbosity', type=click.Choice(['0', '1', '2']),
    default='1', show_default=True, help='Status/progress reporting verbosity')
@click.argument('config_file', type=click.Path(exists=True, dir_okay=False))
@click.argument('dest_path', type=click.Path(exists=True, dir_okay=True))
def apply(overwrite: bool, recurse: bool, verbosity: str,
          config_file: str, dest_path: str):
    '''Applies specified .cfg file to THE400 Mini USB Media games.'''
    extensions = get_extensions()
    target_files = build_target_file_list(pathlib.Path(dest_path), extensions,
                                          recurse)

    if int(verbosity) == PROGRESS:
        # ... showing a progress bar.
        with click.progressbar(target_files, label='Applying config') as bar:
            for file in bar:
                apply_config(config_file, file, overwrite, verbosity)        
    else:
        for file in target_files:
            apply_config(config_file, file, overwrite, verbosity)           
 
    exit(SUCCESS)

@config.command()
@click.option('-r', '--recurse', is_flag=True, default=False,
    help='Recursively process all files in target directory')
@click.option('-v', '--verbosity', type=click.Choice(['0', '1', '2']),
    default='1', show_default=True, help='Status/progress reporting verbosity')
@click.argument('update_file', type=click.Path(exists=True, dir_okay=False))
@click.argument('dest_path', type=click.Path(exists=True, dir_okay=True))
def update(recurse: bool, verbosity: str, update_file: str, dest_path: str):
    '''Updates .cfg files with settings from specified update file.'''
    # Load the update file, ONCE:
    update_file_data = load_config_data(pathlib.Path(update_file))

    # Our target files are all files with the .cfg extension    
    extensions = [CFG_EXTENSION]
    target_files = build_target_file_list(pathlib.Path(dest_path), extensions,
                                          recurse)
    
    if int(verbosity) == PROGRESS:
        # ... showing a progress bar.
        with click.progressbar(target_files, label='Updating config') as bar:
            for target_file in bar:
                update_config_file(update_file_data, target_file)        
    else:
        for target_file in target_files:
            update_config_file(update_file_data, target_file)
            echo_v(f'Updated: {target_file} with: {update_file}',
                   int(verbosity))      

    exit(SUCCESS)

# Configuration File Functions
def build_config(model: str, basic: bool, video_std: str, force_pal: bool,
        artefact: str, start: int, height: int, width:int) -> list:
    entries = []
    add_emulator_settings(entries, model, basic, video_std, force_pal, artefact)
    add_input_analog_settings(entries)    
    add_display_settings(entries, start, height, width)
    add_controller_mappings(entries, model)
    return entries

def add_emulator_settings(config: list, model: str, basic: bool, video_std: str,
    force_pal: bool, artefact: str):
    '''Adds emulator settings to the configuration file.'''
    # Build the main machine/model/basic entry
    model_value = MODEL_MACHINE_ID[model]
    basic_value = f'{EMULATOR_SEPARATOR}{BASIC}' if basic else ''
    machine = (f'{KEY_EMULATOR_PREFIX}{KEY_EMULATOR_MACHINE}'
               f' = "{model_value}{basic_value}'
               f'{EMULATOR_SEPARATOR}{video_std.lower()}"{EOL}')
    config.append(machine)

    # Add artefact and force_pal settings
    config.append(f'{KEY_EMULATOR_PREFIX}{KEY_EMULATOR_ARTEFACT}'
                    f' = {artefact}{EOL}')
    config.append(f'{KEY_EMULATOR_PREFIX}{KEY_EMULATOR_FORCE_PAL}'
                    f' = {str(force_pal).lower()}{EOL}')

def add_input_analog_settings(config: list):
    '''Adds default input_analogue settings to the configuration file.'''
    config.append(f'{KEY_INPUT_ANALOG_PREFIX}{KEY_ANALOG_RATE}'
                  f' = {DEFAULT_ANALOG_RATE}{EOL}')
    config.append(f'{KEY_INPUT_ANALOG_PREFIX}{KEY_ANALOG_CURVE}'
                  f' = "{DEFAULT_ANALOG_CURVE}"{EOL}')
    config.append(f'{KEY_INPUT_ANALOG_PREFIX}{KEY_ANALOG_AUTO_CENTER}'
                  f' = "{DEFAULT_AUTO_CENTER}"{EOL}')

def add_display_settings(config: list, start: int, height: int, width: int):
    '''Adds default display settings to the configuration file.'''
    config.append(f'{KEY_DISPLAY_PREFIX}{KEY_DISPLAY_START}'
                  f' = {start}{EOL}')
    config.append(f'{KEY_DISPLAY_PREFIX}{KEY_DISPLAY_HEIGHT}'
                  f' = {height}{EOL}')
    config.append(f'{KEY_DISPLAY_PREFIX}{KEY_DISPLAY_WIDTH}'
                  f' = {width}{EOL}')
    
def add_controller_mappings(config: list, model: str):
    '''Adds default controller mapping to the configuration file.'''
    # Select the correct set of controller values
    controller_values = (
        CONTROLLER_DEFAULTS if model != '5200' else CONTROLLER_5200_DEFAULTS)
    
    # Add default controller settings
    for controller in range(NUMBER_OF_CONTROLLERS):
        for key in controller_values:
            line = (f'{CONTROLLER_PREIX}{controller}{CONTROLLER_SEPARATOR}{key}'
                    f' = "{controller_values[key]}"{EOL}')
            config.append(line)

    return config

def update_config(config_updates: list, target_config: list):
    '''Updates the specified configuration file with the new settings.'''
    # Iterate through the new configuration settings   
    for update_item in config_updates:
        item_found = False
        update_key = get_config_item_key(update_item)
        for target_item in target_config:
            # Update the existing config item if it exists ...
            if update_key == get_config_item_key(target_item):               
                target_config[target_config.index(target_item)] = update_item              
                item_found = True
                break           
        
        if not item_found:
            # ... otherwise add the new config item.            
            target_config.append(update_item)  
            
def get_config_item_key(config_item: str) -> str:
    '''Extracts the key from a configuration line item.'''
    return config_item.split(KEY_VALUE_SEPARATOR)[0].strip()

def get_extensions() -> list:
    '''Returns the list of supported Atari media file extensions.'''
    # Start with the default list of extensions ...
    extensions = MEDIA_EXTENSIONS
    # ... and add all the specific cartridge extensions:
    for number in range(FIRST_CARTRIDGE_TYPE, LAST_CARTRIDGE_TYPE + 1):
        extensions.append(f'c{number:02d}')

    return extensions

def build_target_file_list(dest_path: pathlib.Path, extensions: list,
                           recurse: bool = False) -> list:
    # Anything files with extensions not in this list are excluded    
    target_files = []
    # Single file?
    if dest_path.is_file() and dest_path.suffix[1:].lower() in extensions:
        # Add file, replacing the extension with '.cfg'
        target_files.append(dest_path.with_suffix(CFG_SUFFIX))
    
    # Directory?
    if dest_path.is_dir():
        # Recurse if specied ...
        pattern = TARGET_PATTERN_RECR if recurse else TARGET_PATTERN  
        # ... and find all files ...      
        candidate_files = list(dest_path.glob(pattern))  
        # ... adding only those with valid extensions, but replacing the
        # extension with '.cfg'      
        target_files = [file.with_suffix(CFG_SUFFIX)
                        for file in candidate_files
                        if (file.is_file() and
                            file.suffix[1:].lower() in extensions)]
    
    return target_files

def apply_config(config_file: str, target_file: pathlib.Path, overwrite: bool,
                 verbosity: str):
    '''Applies the specified configuration file to the target file.'''
    if not target_file.exists() or overwrite:
        shutil.copy(config_file, target_file)
        echo_v(f'Applied {pathlib.Path(config_file)} to: {target_file}',
                int(verbosity))

def load_config_data(update_file: pathlib.Path) -> list:
    '''Loads the specified update file into a list of configuration items.'''    
    with open(update_file, 'r') as file:
        lines = [line.rstrip() for line in file]
    return lines

def update_config_file(update_file: list, target_file: pathlib.Path):
    '''Updates the specified configuration file with the new settings.'''
    target_config = load_config_data(target_file)
    # Update the target file with the new settings
    update_config(update_file, target_config)
    # Write the updated target file
    with open(target_file, 'w') as file:
        for line in target_config:
            file.write(line + '\n')

# GENERAL Utility Functions

def echo_v(message: str, verbosity: int):
    if verbosity == VERBOSE:
        click.echo(message)	

# Run!
if __name__ == '__main__':
    config()


#!python3

# atr.py - Atari 8-bit ATR Disk Image Utility & Functions
#
# Copyright(C) 2026, Ian Michael Dunmore
#
# License: https://github.com/idunmore/aemt/blob/master/LICENSE

# Native Python Modules
import pathlib

# 3rd Party/External Modules
import click

# Local Application Modules
from .style import C

# Constants

# Error Messages and Command Result Exit Codes
ERROR_TEXT = f'{C.Err}Error: {C.Off}'
ERROR = 1
SUCCESS = 0

# Status Messages
PROTECTED = f'{C.Prot}Write-protected{C.Off}'
UNPROTECTED = f'{C.Unprot}Writeable{C.Off}'

# .ATR File Constants
ATR_PATTERN = '*.atr'
ATR_PATTERN_RECR = '**/*.atr'
STATUS_BYTE_INDEX = 0x0F
PROTECT_BIT_MASK = 0x01
UNPROTECT_BIT_MASK = 0xFE

@click.group()
@click.version_option()
def atr():
    '''Manipulates Atari 8-bit .ATR disk images.'''
    pass

@atr.command('protect')
@click.option('-r', '--recurse', is_flag=True, default=False,
    help='Process directories recursively for .atr files')
@click.option('-v', '--verbose', is_flag=True, default=False,
    help='Verbose output')
@click.argument('source_path', 
    type=click.Path(exists=True, file_okay=True, dir_okay=True))
def protect(recurse: bool, verbose: bool, source_path: str):
    '''Write-Protect .ATR disk images.'''
    ret_val = process_atr_files(source_path, recurse, verbose, protect=True)
    exit(ret_val)

@atr.command('unprotect')
@click.option('-r', '--recurse', is_flag=True, default=False,
    help='Process directories recursively for .atr files')
@click.option('-v', '--verbose', is_flag=True, default=False,
    help='Verbose output')
@click.argument('source_path', 
    type=click.Path(exists=True, file_okay=True, dir_okay=True))
def unprotect(recurse: bool, verbose: bool, source_path: str):
    '''Remove write-protection from .ATR disk images.'''
    ret_val = process_atr_files(source_path, recurse, verbose, protect=False)
    exit(ret_val)

@atr.command('status')
@click.option('-r', '--recurse', is_flag=True, default=False,
    help='Process directories recursively for .atr files')
@click.argument('source_path', 
    type=click.Path(exists=True, file_okay=True, dir_okay=True))
def status(recurse: bool, source_path: str):
    '''Display the write-protection status of .ATR disk images.'''
    files = build_source_file_list(source_path, recurse)

    if not files:
        click.echo(ERROR_TEXT + 'No .atr files found to process.')
        exit(ERROR)
    
    for file in files:
        is_protected = get_file_protection_status(file)
        status_text = f'{PROTECTED if is_protected else UNPROTECTED}'
        click.echo(f'[{status_text:>25}] {file}')
        exit(SUCCESS)

def build_source_file_list(source_path: str, recurse: bool) -> list:
    # We can work on a single file, or a directory (with optional recursion),
    # so build a list of file(s) accordingly
    source_path = pathlib.Path(source_path)
    files = None
    if source_path.is_file():
        files = []
        files.append(source_path)
    elif source_path.is_dir():
        pattern = ATR_PATTERN_RECR if recurse else ATR_PATTERN      
        files = list(source_path.glob(pattern))
        files.sort(key=lambda x: x.name.lower())
            
    return files

def process_atr_files(
    source_path: str, recurse: bool, verbose: bool, protect: bool) -> int:

    files = build_source_file_list(source_path, recurse)
    if not files:
        click.echo(ERROR_TEXT + 'No .atr files found to process.')
        return ERROR

    for file in files:
        set_file_protection(file, protect)

        if verbose:
            action = f'[{PROTECTED if protect else UNPROTECTED:>25}]'
            click.echo(f'{action} {file}')

    return SUCCESS

def get_file_protection_status(file: pathlib.Path) -> bool:
    with open(file, 'rb') as atr_file:
        # Read the status byte from the header
        atr_file.seek(STATUS_BYTE_INDEX)
        status_byte = atr_file.read(1)[0]
        atr_file.close()             

    return (status_byte & PROTECT_BIT_MASK) != 0

def set_file_protection(file: pathlib.Path, protect: bool) -> int:
    with open(file, 'rb+' ) as atr_file:
        # Read the status byte from the header
        atr_file.seek(STATUS_BYTE_INDEX)
        status_byte = atr_file.read(1)[0]

        # Modify the status byte to set or clear the write-protect bit
        new_status_byte = set_protect_bit(bytes([status_byte]), protect)[0]
       
        # Write the modified status byte back to the header
        atr_file.seek(STATUS_BYTE_INDEX)
        atr_file.write(bytes([new_status_byte]))
        
        atr_file.flush()
        atr_file.close()

    return SUCCESS

def set_protect_bit(status_byte: bytes, protect: bool) -> bytes:
    if protect:
        return bytes([status_byte[0] | PROTECT_BIT_MASK])
    else:
        return bytes([status_byte[0] & UNPROTECT_BIT_MASK])    

# Run!
if __name__ == '__main__':
    atr()


# style.py - Style Classes, Constants and Functions for AEMT
#
# Copyright(C) 2026, Ian Michael Dunmore
#
# License: https://github.com/idunmore/aemt/blob/master/LICENSE

# 3rd Party/External Modules
from sty import fg

class C:
    """
    Text color/effect aliases (from "sty" module values).
    Use with f-strings: "{C.Err}Error text.{C.Off}
    """
    # C class = "Color"; Abbreviations are to reduce line lengths.
    Err = fg.red        # For Errors
    Prot = fg.green     # For Write-Protect 
    Unprot = fg.red     # For Writeable (Unprotected)
    Valid = fg.green    # For Valid/Verified
    Invalid = fg.red    # For Invalid/Failed Verification
    Off = fg.rs         # Default Color (Color="OFF")



#!python3

# cartridge.py - Atari 8-bit Cartridge Utility & Functions
#
# Copyright(C) 2026, Ian Michael Dunmore
#
# License: https://github.com/idunmore/aemt/blob/master/LICENSE

# Native Python Modules

import enum
import pathlib
from typing import Self

# 3rd Party/External Modules
import click

# Local Application Modules
from .style import C

# Constants

# Error Messages and Command Result Exit Codes
ERROR_TEXT = f'{C.Err}Error: {C.Off}'
ERROR = 1
SUCCESS = 0

# Verbosity Values
SILENT = 0
PROGRESS = 1
VERBOSE = 2

# General Constants
BYTES_PER_KILOBYTE = 1024

# MSB (Most Significant Bit) Constants
MSB_DIGIT_1 = 256 * 256 * 256
MSB_DIGIT_2 = 256 * 256
MSB_DIGIT_3 = 256
MSB_DIGIT_4 = 1

# Cartridge Constants
CART_PATTERN = '*.car'
CART_PATTERN_RECR = '**/*.car'
CHECKSUM_MASK = 0x000000FF
FIRST_CARTRIDGE_TYPE = 1
LAST_CARTRIDGE_TYPE = 70
CART_HEADER_SIZE = 16
CART_PREAMBLE = 'CART'

SIG_OFFSET = 0
SIG_LENGTH = 4
TYPE_OFFSET = 4
TYPE_LENGTH = 4
CHECKSUM_OFFSET = 8
CHECKSUM_LENGTH = 4
IMAGE_OFFSET = 16

# Cartridge (In)validity Reasons
INVALID_SIGNATURE = f'{C.Invalid}Invalid signature{C.Off}'
INVALID_TYPE = f'{C.Invalid}Invalid or unknown cartridge type{C.Off}'
INVALID_CHECKSUM = f'{C.Invalid}Header checksum mismatch with cartridge{C.Off}'
INVALID_SIZE = f'{C.Invalid}Invalid cartridge size{C.Off}'
INVALID_HEADER_SIZE = f'{C.Invalid}Invalid cartridge header size; expected{C.Off}'
VALID = f'{C.Valid}Valid{C.Off}'

# Format Constants
CSV_SEPARATOR = ','
CSV_QUOTE = '"'
SEPARATOR = ' | '
QUOTE = ''
CSV_SIG_START =''
CSV_SIG_END = ''
SIG_START ='['
SIG_END=']'
CSV_TYPE_WIDTH = 1
TYPE_WIDTH = 2
CSV_DESCRIPTION_WIDTH = 1
DESCRIPTION_WIDTH = 46


# Machines Types
class Machine(enum.Enum):
    '''Atari 8-bit Cartridge Machine Types'''
    ATARI_800_XL_XE = 0
    '''Atari 400, 800, 800XL, 1200XL, 600XL, 65XE, 130XE, 800XE, XEGS, and
    compatible machines.'''
    ATARI_800 = 1
    '''Atari 800-Specific - Typically RIGHT SLOT cartridges.'''
    ATARI_5200 = 2

# Cartridge Type Class
class CartridgeType:
    '''Cartridge Type & Details'''
    def __init__(self: Self, type: int, size_kilobytes: int,
        machine: Machine, description: str):        

        if type < FIRST_CARTRIDGE_TYPE or type > LAST_CARTRIDGE_TYPE:
            raise ValueError(f'{INVALID_TYPE}: {type}')
        self._type = type

        if (size_kilobytes > BYTES_PER_KILOBYTE and
            size_kilobytes % BYTES_PER_KILOBYTE != 0):
            raise ValueError(f'{INVALID_SIZE}: {size_kilobytes}')        
        self._size_kilobytes = size_kilobytes

        self._machine = machine
        self._description = description        

    @property
    def type(self: Self) -> int:
        return self._type
    
    @property
    def size_kilobytes(self: Self) -> int:
        return self._size_kilobytes
    
    @property
    def machine(self: Self) -> str:
        return self._machine
    
    @property
    def description(self: Self) -> str:
        return self._description
    
# Currently (04/19/24) known cartridge ID/types
cart_types = {
    1: CartridgeType(1, 8,  Machine.ATARI_800_XL_XE, 'Standard 8 KB cartridge', ),
    2: CartridgeType(2, 16, Machine.ATARI_800_XL_XE, 'Standard 16 KB cartridge'),
    3: CartridgeType(3, 16, Machine.ATARI_800_XL_XE, 'OSS two-chip 16 KB cartridge'),        
    4: CartridgeType(4, 32, Machine.ATARI_5200, 'Standard 32 KB 5200 cartridge'),
    5: CartridgeType(5, 32, Machine.ATARI_800_XL_XE, 'DB 32 KB cartridge'),
    6: CartridgeType(6, 16, Machine.ATARI_5200, 'Two chip 16 KB 5200 cartridge'),
    7: CartridgeType(7, 40, Machine.ATARI_5200, 'Bounty Bob Strikes Back 40 KB 5200 cartridge'),   
    8: CartridgeType(8, 64, Machine.ATARI_800_XL_XE, '64 KB Williams cartridge'),
    9: CartridgeType(9, 64, Machine.ATARI_800_XL_XE, 'Express 64 KB cartridge'),
    10: CartridgeType(10, 64, Machine.ATARI_800_XL_XE, 'Diamond 64 KB cartridge'),
    11: CartridgeType(11, 64, Machine.ATARI_800_XL_XE, 'SpartaDOS X 64 KB cartridge'),
    12: CartridgeType(12, 32, Machine.ATARI_800_XL_XE, 'XEGS 32 KB cartridge'),
    13: CartridgeType(13, 64, Machine.ATARI_800_XL_XE, 'XEGS 64 KB cartridge (banks 0-7)'),
    14: CartridgeType(14, 128, Machine.ATARI_800_XL_XE, 'XEGS 128 KB cartridge'),
    15: CartridgeType(15, 16, Machine.ATARI_800_XL_XE, 'OSS one chip 16 KB cartridge'),
    16: CartridgeType(16, 16, Machine.ATARI_5200, 'One chip 16 KB 5200 cartridge'),
    17: CartridgeType(17, 128, Machine.ATARI_800_XL_XE, 'Decoded Atrax 128 KB cartridge'),
    18: CartridgeType(18, 40, Machine.ATARI_800_XL_XE, 'Bounty Bob Strikes Back 40 KB cartridge'),
    19: CartridgeType(19, 8, Machine.ATARI_5200, 'Standard 8 KB 5200 cartridge'),
    20: CartridgeType(20, 4, Machine.ATARI_5200, 'Standard 4 KB 5200 cartridge'),
    21: CartridgeType(21, 8, Machine.ATARI_800, 'Right slot 8 KB cartridge'),
    22: CartridgeType(22, 32, Machine.ATARI_800_XL_XE, '32 KB Williams cartridge'),
    23: CartridgeType(23, 256, Machine.ATARI_800_XL_XE, 'XEGS 256 KB cartridge'),
    24: CartridgeType(24, 512, Machine.ATARI_800_XL_XE, 'XEGS 512 KB cartridge'),
    25: CartridgeType(25, 1024, Machine.ATARI_800_XL_XE, 'XEGS 1 MB cartridge'),
    26: CartridgeType(26, 16, Machine.ATARI_800_XL_XE, 'MegaCart 16 KB cartridge'),
    27: CartridgeType(27, 32, Machine.ATARI_800_XL_XE, 'MegaCart 32 KB cartridge'),
    28: CartridgeType(28, 64, Machine.ATARI_800_XL_XE, 'MegaCart 64 KB cartridge'),
    29: CartridgeType(29, 128, Machine.ATARI_800_XL_XE, 'MegaCart 128 KB cartridge'),
    30: CartridgeType(30, 256, Machine.ATARI_800_XL_XE, 'MegaCart 256 KB cartridge'),
    31: CartridgeType(31, 512, Machine.ATARI_800_XL_XE, 'MegaCart 512 KB cartridge'),
    32: CartridgeType(32, 1024, Machine.ATARI_800_XL_XE, 'MegaCart 1 MB cartridge'),
    33: CartridgeType(33, 32, Machine.ATARI_800_XL_XE, 'Switchable XEGS 32 KB cartridge'),
    34: CartridgeType(34, 64, Machine.ATARI_800_XL_XE, 'Switchable XEGS 64 KB cartridge'),
    35: CartridgeType(35, 128, Machine.ATARI_800_XL_XE, 'Switchable XEGS 128 KB cartridge'),
    36: CartridgeType(36, 256, Machine.ATARI_800_XL_XE, 'Switchable XEGS 256 KB cartridge'),
    37: CartridgeType(37, 512, Machine.ATARI_800_XL_XE, 'Switchable XEGS 512 KB cartridge'),
    38: CartridgeType(38, 1024, Machine.ATARI_800_XL_XE, 'Switchable XEGS 1 MB cartridge'),
    39: CartridgeType(39, 8, Machine.ATARI_800_XL_XE, 'Phoenix 8 KB cartridge'),
    40: CartridgeType(40, 16, Machine.ATARI_800_XL_XE, 'Blizzard 16 KB cartridge'),
    41: CartridgeType(41, 128, Machine.ATARI_800_XL_XE, 'Atarimax 128 KB Flash cartridge'),
    42: CartridgeType(42, 1024, Machine.ATARI_800_XL_XE, 'Atarimax 1 MB Flash cartridge'),
    43: CartridgeType(43, 128, Machine.ATARI_800_XL_XE, 'SpartaDOS X 128 KB cartridge'),
    44: CartridgeType(44, 8, Machine.ATARI_800_XL_XE, 'OSS 8 KB cartridge'),
    45: CartridgeType(45, 16, Machine.ATARI_800_XL_XE, 'OSS two chip 16 KB cartridge (043M)'),
    46: CartridgeType(46, 4, Machine.ATARI_800_XL_XE, 'Blizzard 4 KB cartridge'),
    47: CartridgeType(47, 32, Machine.ATARI_800_XL_XE, 'AST 32 KB cartridge'),
    48: CartridgeType(48, 64, Machine.ATARI_800_XL_XE, 'Atrax SDX 64 KB cartridge'),
    49: CartridgeType(49, 128, Machine.ATARI_800_XL_XE, 'Atrax SDX 128 KB cartridge'),
    50: CartridgeType(50, 64, Machine.ATARI_800_XL_XE, 'Turbosoft 64 KB cartridge'),
    51: CartridgeType(51, 128, Machine.ATARI_800_XL_XE, 'Turbosoft 128 KB cartridge'),
    52: CartridgeType(52, 32, Machine.ATARI_800_XL_XE, 'Ultracart 32 KB cartridge'),
    53: CartridgeType(53, 8, Machine.ATARI_800_XL_XE, 'Low bank 8 KB cartridge'),
    54: CartridgeType(54, 128, Machine.ATARI_800_XL_XE, 'SIC! 128 KB cartridge'),
    55: CartridgeType(55, 256, Machine.ATARI_800_XL_XE, 'SIC! 256 KB cartridge'),
    56: CartridgeType(56, 512, Machine.ATARI_800_XL_XE, 'SIC! 512 KB cartridge'),
    57: CartridgeType(57, 2, Machine.ATARI_800_XL_XE, 'Standard 2 KB cartridge'),
    58: CartridgeType(58, 4, Machine.ATARI_800_XL_XE, 'Standard 4 KB cartridge'),
    59: CartridgeType(59, 4, Machine.ATARI_800, 'Right slot 4 KB cartridge'),
    60: CartridgeType(60, 32, Machine.ATARI_800_XL_XE, 'Blizzard 32 KB cartridge'),
    61: CartridgeType(61, 2048, Machine.ATARI_800_XL_XE, 'MegaMax 2 MB cartridge'),
    62: CartridgeType(62, 128*1024, Machine.ATARI_800_XL_XE, 'The!Cart 128 MB cartridge'),
    63: CartridgeType(63, 4096, Machine.ATARI_800_XL_XE, 'Flash MegaCart 4 MB cartridge'),
    64: CartridgeType(64, 2048, Machine.ATARI_800_XL_XE, 'MegaCart 2 MB cartridge'),
    65: CartridgeType(65, 32*1024, Machine.ATARI_800_XL_XE, 'The!Cart 32 MB cartridge'),
    66: CartridgeType(66, 64*1024, Machine.ATARI_800_XL_XE, 'The!Cart 64 MB cartridge'),
    67: CartridgeType(67, 64, Machine.ATARI_800_XL_XE, 'XEGS 64 KB cartridge (banks 8-15)'),
    68: CartridgeType(68, 128, Machine.ATARI_800_XL_XE, 'Atrax 128 KB cartridge'),
    69: CartridgeType(69, 32, Machine.ATARI_800_XL_XE, 'aDawliah 32 KB cartridge'),
    70: CartridgeType(70, 64, Machine.ATARI_800_XL_XE, 'aDawliah 64 KB cartridge')
}

class CartridgeHeader:
    def __init__(self: Self, bytes: bytes):
        if (len(bytes)) < CART_HEADER_SIZE:
            raise ValueError(
                f'{INVALID_HEADER_SIZE}: {CART_HEADER_SIZE} '
                f'bytes, got: {len(bytes)}')
        self._raw_bytes = bytes
        self._invalid_reason = None

    @property
    def signature(self: Self) -> str:
        sig = self._raw_bytes[SIG_OFFSET:SIG_LENGTH]
        return f'{chr(sig[0])}{chr(sig[1])}{chr(sig[2])}{chr(sig[3])}'
    
    @property
    def type(self: Self) -> int:
        type = self._raw_bytes[TYPE_OFFSET:TYPE_OFFSET + TYPE_LENGTH]
        return ((type[0] * MSB_DIGIT_1) + (type[1] * MSB_DIGIT_2) +
                (type[2] * MSB_DIGIT_3) + (type[3] * MSB_DIGIT_4))

    @property
    def checksum(self: Self) -> int:
        checksum = (
            self._raw_bytes[CHECKSUM_OFFSET:CHECKSUM_OFFSET + CHECKSUM_LENGTH])
        return ((checksum[0] * MSB_DIGIT_1) + (checksum[1] * MSB_DIGIT_2) +
                (checksum[2] * MSB_DIGIT_3) + (checksum[3] * MSB_DIGIT_4))        
    
    @property
    def description(self: Self) -> str:
        if self.type in cart_types:
            return cart_types[self.type].description
        else:
            return INVALID_TYPE
    
    @property
    def is_valid(self: Self) -> bool:
        if self.signature != CART_PREAMBLE:
            self._invalid_reason = INVALID_SIGNATURE
            return False
        if self.type not in cart_types:
            self._invalid_reason = INVALID_TYPE
            return False
        if self.checksum != compute_checksum(self._raw_bytes[IMAGE_OFFSET:]):
            self._invalid_reason = INVALID_CHECKSUM
            return False
        
        self._invalid_reason = VALID
        return True
    
    @property
    def invalid_reason(self: Self) -> str:
        '''Reason cartridge is invalid, if any; otherwise None'''
        # Invoke "is_valid" to set the invalid reason
        is_valid = self.is_valid
        return self._invalid_reason

# General Atari Functions
def compute_checksum(bytes: bytes) -> int:
    '''Computes the .car checksum for a sequence of bytes'''
    result = 0
    for byte in bytes:
        result += byte & CHECKSUM_MASK
    return result

@click.group()
@click.version_option()
def cart():
    '''Identifies and validates Atari 8-bit cartridges.'''
    pass

@cart.command('list_types')
def list_types():
    '''Lists known cartridge types and specifications.'''
    print('Type | Size (KB)| Machine(s) | Description')
    print('-----|----------|------------|' + '-' * 45)
    for cart in cart_types.values():        
        print(f' {cart.type:3}{SEPARATOR}{cart.size_kilobytes:8,}{SEPARATOR}'
              f'{str(cart.machine)[14:]:<10}{SEPARATOR}{cart.description}')
    exit(SUCCESS)

@cart.command('id')
@click.option('-c', '--csv', is_flag=True, default=False,
    help='Output in CSV format')
@click.option('-h', '--header', is_flag=True, default=False,
    help='Output a header if in CSV format')
@click.option('-r', '--recurse', is_flag=True, default=False,
    help='Process directories recursively for .car files')
@click.option('-v', '--verbose', is_flag=True, default=False,
    help='Verbose output')
@click.argument('source_path', 
    type=click.Path(exists=True, file_okay=True, dir_okay=True))
def id(csv: bool, header: bool, recurse: bool, verbose: bool, source_path: str):
    '''Identifies the cartridge type and verifies header and data.

    \b
      SOURCE_PATH may be a directory or a file; if a directory *only* .car files
      will be processed.  The -r/--recurse option will include subdirectories.
    '''
    files = build_source_file_list(source_path, recurse)
    if files is None:    
        print(f'No files to identify.')
        exit(SUCCESS)

    if csv and header:
        if verbose:
            print(f'Signature,Type,Checksum,Actual Checksum,Is Valid,'
                  f'Description,Reason,File')
        else:
            print(f'Signature,Type,Checksum,Actual Checksum,Is Valid')
    
    for file in files:
        id_cartridge(file, csv, verbose)
    
    exit(SUCCESS)

def build_source_file_list(source_path: str, recurse: bool) -> list:
    # We can work on a single file, or a directory (with optional recursion),
    # so build a list of file(s) accordingly
    source_path = pathlib.Path(source_path)
    files = None
    if source_path.is_file():
        files = []
        files.append(source_path)
    elif source_path.is_dir():
        pattern = CART_PATTERN_RECR if recurse else CART_PATTERN      
        files = list(source_path.glob(pattern))
        files.sort(key=lambda x: x.name.lower())
            
    return files

def id_cartridge(file: pathlib.Path, csv: bool, verbose: bool):
    # Get the cartridge header and validate it
    data = file.read_bytes()
    header = CartridgeHeader(data)
    actual_checksum = compute_checksum(data[CART_HEADER_SIZE:])

    # Setup output formatting
    sep, quote = (CSV_SEPARATOR, CSV_QUOTE) if csv else (SEPARATOR, QUOTE)
    sig_start, sig_end = ((CSV_SIG_START, CSV_SIG_END) if csv
                           else (SIG_START, SIG_END))
    cart_type_width = CSV_TYPE_WIDTH if csv else TYPE_WIDTH
    cart_description_width = CSV_DESCRIPTION_WIDTH if csv else DESCRIPTION_WIDTH
    
    # Build the base output ...    
    item = (f'{sig_start}{header.signature}{sig_end}{sep}'
            f'{header.type:{cart_type_width}}{sep}0x{header.checksum:08x}{sep}'
            f'0x{actual_checksum:08x}{sep}'            
            f'{header.is_valid}{sep}'
            f'{quote}{header.description:{cart_description_width}}{quote}')
    
    # ... and add the verbose elements if requested
    if verbose:
        item += (f'{sep}{quote}{header.invalid_reason}{quote}{sep}'
                 f'{quote}{str(file)}{quote}')              
    
    print(item)

# Run!
if __name__ == '__main__':
    cart()


#!python3

# AEMT.py - [A]tari [E]ight-bit [M]ulti-[T]ool
#
# Copyright(C) 2026, Ian Michael Dunmore
#
# License: https://github.com/idunmore/aemt/blob/master/LICENSE

# 3rd Party/External Modules
import sys
import click

# Local Application Modules
from .atr import atr
from .cartridge import cart
from .config import config
from .split import split

# Constants

# Error Messages and Command Result Exit Codes
ERROR_TEXT ='Error: '
ERROR = 1
SUCCESS = 0

# Verbosity Values
SILENT = 0
PROGRESS = 1
VERBOSE = 2

# Command Line Interface

@click.group()
@click.version_option()
def aemt():
    '''[A]tari [E]ight-bit [M]ulti-[T]ool

            \b
                Manipulates Atari 8-bit .ATR disk images.                        
            
            \b
                Moves files to organized folder structures, with an optionally
                limited number of files per folder.
            
            \b           
                Creates, applies and updates .cfg files for Atari THE400 Mini
                games on USB media.
            
            \b
                Identifies and verifies Atari 8-bit cartridge images.'''
    pass
    
# Run!
if __name__ == 'aemt.cli':
    aemt.add_command(atr)
    aemt.add_command(cart)
    aemt.add_command(config)
    aemt.add_command(split)
    aemt()

if __name__ == '__main__':
    sys.exit(aemt())
